@@ -29,27 +29,35 @@
 /**************************************************************************/
 
 #include "import_dock.h"
-#include "editor_node.h"
-#include "editor_resource_preview.h"
-#include "editor_scale.h"
+
+#include "core/config/project_settings.h"
+#include "editor/editor_node.h"
+#include "editor/editor_resource_preview.h"
+#include "editor/editor_settings.h"
+#include "editor/editor_string_names.h"
+#include "editor/editor_undo_redo_manager.h"
+#include "editor/themes/editor_scale.h"
+#include "editor/themes/editor_theme_manager.h"
 
 class ImportDockParameters : public Object {
 	GDCLASS(ImportDockParameters, Object);
 
 public:
-	Map<StringName, Variant> values;
+	HashMap<StringName, Variant> values;
 	List<PropertyInfo> properties;
 	Ref<ResourceImporter> importer;
 	Vector<String> paths;
-	Set<StringName> checked;
-	bool checking;
+	HashSet<StringName> checked;
+	bool checking = false;
+	bool skip = false;
+	String base_options_path;
 
 	bool _set(const StringName &p_name, const Variant &p_value) {
 		if (values.has(p_name)) {
 			values[p_name] = p_value;
 			if (checking) {
 				checked.insert(p_name);
-				_change_notify();
+				notify_property_list_changed();
 			}
 			return true;
 		}
@@ -66,14 +74,14 @@ public:
 		return false;
 	}
 	void _get_property_list(List<PropertyInfo> *p_list) const {
-		for (const List<PropertyInfo>::Element *E = properties.front(); E; E = E->next()) {
-			if (!importer->get_option_visibility(E->get().name, values)) {
+		for (const PropertyInfo &E : properties) {
+			if (!importer->get_option_visibility(base_options_path, E.name, values)) {
 				continue;
 			}
-			PropertyInfo pi = E->get();
+			PropertyInfo pi = E;
 			if (checking) {
 				pi.usage |= PROPERTY_USAGE_CHECKABLE;
-				if (checked.has(E->get().name)) {
+				if (checked.has(E.name)) {
 					pi.usage |= PROPERTY_USAGE_CHECKED;
 				}
 			}
@@ -82,17 +90,15 @@ public:
 	}
 
 	void update() {
-		_change_notify();
-	}
-
-	ImportDockParameters() {
-		checking = false;
+		notify_property_list_changed();
 	}
 };
 
+ImportDock *ImportDock::singleton = nullptr;
+
 void ImportDock::set_edit_path(const String &p_path) {
 	Ref<ConfigFile> config;
-	config.instance();
+	config.instantiate();
 	Error err = config->load(p_path + ".import");
 	if (err != OK) {
 		clear();
@@ -100,30 +106,39 @@ void ImportDock::set_edit_path(const String &p_path) {
 	}
 
 	String importer_name = config->get_value("remap", "importer");
-
-	params->importer = ResourceFormatImporter::get_singleton()->get_importer_by_name(importer_name);
+	if (importer_name == "keep") {
+		params->importer.unref();
+		params->skip = false;
+	} else if (importer_name == "skip") {
+		params->importer.unref();
+		params->skip = true;
+	} else {
+		params->importer = ResourceFormatImporter::get_singleton()->get_importer_by_name(importer_name);
+		params->skip = false;
+	}
 
 	params->paths.clear();
 	params->paths.push_back(p_path);
+	params->base_options_path = p_path;
 
-	_update_options(config);
+	_update_options(p_path, config);
 
 	List<Ref<ResourceImporter>> importers;
 	ResourceFormatImporter::get_singleton()->get_importers_for_extension(p_path.get_extension(), &importers);
 	List<Pair<String, String>> importer_names;
 
-	for (List<Ref<ResourceImporter>>::Element *E = importers.front(); E; E = E->next()) {
-		importer_names.push_back(Pair<String, String>(E->get()->get_visible_name(), E->get()->get_importer_name()));
+	for (const Ref<ResourceImporter> &E : importers) {
+		importer_names.push_back(Pair<String, String>(E->get_visible_name(), E->get_importer_name()));
 	}
 
 	importer_names.sort_custom<PairSort<String, String>>();
 
 	import_as->clear();
 
-	for (List<Pair<String, String>>::Element *E = importer_names.front(); E; E = E->next()) {
-		import_as->add_item(E->get().first);
-		import_as->set_item_metadata(import_as->get_item_count() - 1, E->get().second);
-		if (E->get().second == importer_name) {
+	for (const Pair<String, String> &E : importer_names) {
+		import_as->add_item(E.first);
+		import_as->set_item_metadata(-1, E.second);
+		if (E.second == importer_name) {
 			import_as->select(import_as->get_item_count() - 1);
 		}
 	}
@@ -142,52 +157,97 @@ void ImportDock::set_edit_path(const String &p_path) {
 
 void ImportDock::_add_keep_import_option(const String &p_importer_name) {
 	import_as->add_separator();
-	import_as->add_item(TTR("Keep File (No Import)"));
-	import_as->set_item_metadata(import_as->get_item_count() - 1, "keep");
+	import_as->add_item(TTR("Keep File (exported as is)"));
+	import_as->set_item_metadata(-1, "keep");
+	import_as->add_item(TTR("Skip File (not exported)"));
+	import_as->set_item_metadata(-1, "skip");
 	if (p_importer_name == "keep") {
+		import_as->select(import_as->get_item_count() - 2);
+	} else if (p_importer_name == "skip") {
 		import_as->select(import_as->get_item_count() - 1);
 	}
 }
 
-void ImportDock::_update_options(const Ref<ConfigFile> &p_config) {
+void ImportDock::_update_options(const String &p_path, const Ref<ConfigFile> &p_config) {
+	// Set the importer class to fetch the correct class in the XML class reference.
+	// This allows tooltips to display when hovering properties.
+	if (params->importer.is_valid()) {
+		// Null check to avoid crashing if the "Keep File (exported as is)" mode is selected.
+		import_opts->set_object_class(params->importer->get_class_name());
+	}
+
 	List<ResourceImporter::ImportOption> options;
 
 	if (params->importer.is_valid()) {
-		params->importer->get_import_options(&options);
+		params->importer->get_import_options(p_path, &options);
 	}
 
 	params->properties.clear();
 	params->values.clear();
 	params->checking = params->paths.size() > 1;
 	params->checked.clear();
+	params->base_options_path = p_path;
 
-	for (List<ResourceImporter::ImportOption>::Element *E = options.front(); E; E = E->next()) {
-		params->properties.push_back(E->get().option);
-		if (p_config.is_valid() && p_config->has_section_key("params", E->get().option.name)) {
-			params->values[E->get().option.name] = p_config->get_value("params", E->get().option.name);
+	HashMap<StringName, Variant> import_options;
+	if (p_config.is_valid() && p_config->has_section("params")) {
+		List<String> section_keys;
+		p_config->get_section_keys("params", &section_keys);
+		for (const String &section_key : section_keys) {
+			import_options[section_key] = p_config->get_value("params", section_key);
+		}
+		if (params->importer.is_valid()) {
+			params->importer->handle_compatibility_options(import_options);
+		}
+	}
+
+	for (const ResourceImporter::ImportOption &E : options) {
+		params->properties.push_back(E.option);
+		if (p_config.is_valid() && import_options.has(E.option.name)) {
+			params->values[E.option.name] = import_options[E.option.name];
 		} else {
-			params->values[E->get().option.name] = E->get().default_value;
+			params->values[E.option.name] = E.default_value;
 		}
 	}
 
 	params->update();
 	_update_preset_menu();
+
+	bool was_imported = p_config.is_valid() && p_config->get_value("remap", "importer") != "skip" && p_config->get_value("remap", "importer") != "keep";
+	if (was_imported && params->importer.is_valid() && params->paths.size() == 1 && params->importer->has_advanced_options()) {
+		advanced->show();
+		advanced_spacer->show();
+	} else {
+		advanced->hide();
+		advanced_spacer->hide();
+	}
 }
 
 void ImportDock::set_edit_multiple_paths(const Vector<String> &p_paths) {
 	clear();
 
 	// Use the value that is repeated the most.
-	Map<String, Dictionary> value_frequency;
+	HashMap<String, Dictionary> value_frequency;
+	HashSet<String> extensions;
 
 	for (int i = 0; i < p_paths.size(); i++) {
 		Ref<ConfigFile> config;
-		config.instance();
+		config.instantiate();
+		extensions.insert(p_paths[i].get_extension());
 		Error err = config->load(p_paths[i] + ".import");
 		ERR_CONTINUE(err != OK);
 
 		if (i == 0) {
-			params->importer = ResourceFormatImporter::get_singleton()->get_importer_by_name(config->get_value("remap", "importer"));
+			String importer_name = config->get_value("remap", "importer");
+			if (importer_name == "keep") {
+				params->importer.unref();
+				params->skip = false;
+			} else if (importer_name == "skip") {
+				params->importer.unref();
+				params->skip = true;
+			} else {
+				params->importer = ResourceFormatImporter::get_singleton()->get_importer_by_name(importer_name);
+				params->skip = false;
+			}
 			if (params->importer.is_null()) {
 				clear();
 				return;
@@ -201,50 +261,55 @@ void ImportDock::set_edit_multiple_paths(const Vector<String> &p_paths) {
 		List<String> keys;
 		config->get_section_keys("params", &keys);
 
-		for (List<String>::Element *E = keys.front(); E; E = E->next()) {
-			if (!value_frequency.has(E->get())) {
-				value_frequency[E->get()] = Dictionary();
+		for (const String &E : keys) {
+			if (!value_frequency.has(E)) {
+				value_frequency[E] = Dictionary();
 			}
 
-			Variant value = config->get_value("params", E->get());
+			Variant value = config->get_value("params", E);
 
-			if (value_frequency[E->get()].has(value)) {
-				value_frequency[E->get()][value] = int(value_frequency[E->get()][value]) + 1;
+			if (value_frequency[E].has(value)) {
+				value_frequency[E][value] = int(value_frequency[E][value]) + 1;
 			} else {
-				value_frequency[E->get()][value] = 1;
+				value_frequency[E][value] = 1;
 			}
 		}
 	}
 
 	ERR_FAIL_COND(params->importer.is_null());
 
+	String base_path;
+	if (extensions.size() == 1 && p_paths.size() > 0) {
+		base_path = p_paths[0];
+	}
 	List<ResourceImporter::ImportOption> options;
-	params->importer->get_import_options(&options);
+	params->importer->get_import_options(base_path, &options);
 
 	params->properties.clear();
 	params->values.clear();
 	params->checking = true;
 	params->checked.clear();
+	params->base_options_path = base_path;
 
-	for (List<ResourceImporter::ImportOption>::Element *E = options.front(); E; E = E->next()) {
-		params->properties.push_back(E->get().option);
+	for (const ResourceImporter::ImportOption &E : options) {
+		params->properties.push_back(E.option);
 
-		if (value_frequency.has(E->get().option.name)) {
-			Dictionary d = value_frequency[E->get().option.name];
+		if (value_frequency.has(E.option.name)) {
+			Dictionary d = value_frequency[E.option.name];
 			int freq = 0;
 			List<Variant> v;
 			d.get_key_list(&v);
 			Variant value;
-			for (List<Variant>::Element *F = v.front(); F; F = F->next()) {
-				int f = d[F->get()];
+			for (const Variant &F : v) {
+				int f = d[F];
 				if (f > freq) {
-					value = F->get();
+					value = F;
 				}
 			}
 
-			params->values[E->get().option.name] = value;
+			params->values[E.option.name] = value;
 		} else {
-			params->values[E->get().option.name] = E->get().default_value;
+			params->values[E.option.name] = E.default_value;
 		}
 	}
 
@@ -254,18 +319,18 @@ void ImportDock::set_edit_multiple_paths(const Vector<String> &p_paths) {
 	ResourceFormatImporter::get_singleton()->get_importers_for_extension(p_paths[0].get_extension(), &importers);
 	List<Pair<String, String>> importer_names;
 
-	for (List<Ref<ResourceImporter>>::Element *E = importers.front(); E; E = E->next()) {
-		importer_names.push_back(Pair<String, String>(E->get()->get_visible_name(), E->get()->get_importer_name()));
+	for (const Ref<ResourceImporter> &E : importers) {
+		importer_names.push_back(Pair<String, String>(E->get_visible_name(), E->get_importer_name()));
 	}
 
 	importer_names.sort_custom<PairSort<String, String>>();
 
 	import_as->clear();
 
-	for (List<Pair<String, String>>::Element *E = importer_names.front(); E; E = E->next()) {
-		import_as->add_item(E->get().first);
-		import_as->set_item_metadata(import_as->get_item_count() - 1, E->get().second);
-		if (E->get().second == params->importer->get_importer_name()) {
+	for (const Pair<String, String> &E : importer_names) {
+		import_as->add_item(E.first);
+		import_as->set_item_metadata(-1, E.second);
+		if (E.second == params->importer->get_importer_name()) {
 			import_as->select(import_as->get_item_count() - 1);
 		}
 	}
@@ -283,6 +348,29 @@ void ImportDock::set_edit_multiple_paths(const Vector<String> &p_paths) {
 	select_a_resource->hide();
 
 	imported->set_text(vformat(TTR("%d Files"), p_paths.size()));
+
+	if (params->paths.size() == 1 && params->importer->has_advanced_options()) {
+		advanced->show();
+		advanced_spacer->show();
+	} else {
+		advanced->hide();
+		advanced_spacer->hide();
+	}
+}
+
+void ImportDock::reimport_resources(const Vector<String> &p_paths) {
+	switch (p_paths.size()) {
+		case 0:
+			ERR_FAIL_MSG("You need to select files to reimport them.");
+		case 1:
+			set_edit_path(p_paths[0]);
+			break;
+		default:
+			set_edit_multiple_paths(p_paths);
+			break;
+	}
+
+	_reimport_attempt();
 }
 
 void ImportDock::_update_preset_menu() {
@@ -316,75 +404,89 @@ void ImportDock::_importer_selected(int i_idx) {
 	String name = import_as->get_selected_metadata();
 	if (name == "keep") {
 		params->importer.unref();
-		_update_options(Ref<ConfigFile>());
+		params->skip = false;
+		_update_options(params->base_options_path, Ref<ConfigFile>());
+	} else if (name == "skip") {
+		params->importer.unref();
+		params->skip = true;
+		_update_options(params->base_options_path, Ref<ConfigFile>());
 	} else {
 		Ref<ResourceImporter> importer = ResourceFormatImporter::get_singleton()->get_importer_by_name(name);
 		ERR_FAIL_COND(importer.is_null());
 
 		params->importer = importer;
-
+		params->skip = false;
 		Ref<ConfigFile> config;
 		if (params->paths.size()) {
-			config.instance();
-			Error err = config->load(params->paths[0] + ".import");
+			String path = params->paths[0];
+			config.instantiate();
+			Error err = config->load(path + ".import");
 			if (err != OK) {
 				config.unref();
 			}
 		}
-		_update_options(config);
+		_update_options(params->base_options_path, config);
 	}
 }
 
 void ImportDock::_preset_selected(int p_idx) {
 	int item_id = preset->get_popup()->get_item_id(p_idx);
+	String setting_name = "importer_defaults/" + params->importer->get_importer_name();
 
 	switch (item_id) {
 		case ITEM_SET_AS_DEFAULT: {
-			Dictionary d;
-
-			for (const List<PropertyInfo>::Element *E = params->properties.front(); E; E = E->next()) {
-				d[E->get().name] = params->values[E->get().name];
+			Dictionary import_settings;
+			// When import settings already exist, we will update these settings
+			// to ensure that the dictionary retains settings that are not displayed in the
+			// editor. For Scene, the dictionary is the same for FBX, GLTF, and Blender, but each
+			// file type has some different settings.
+			if (ProjectSettings::get_singleton()->has_setting(setting_name)) {
+				import_settings = GLOBAL_GET(setting_name);
 			}
 
-			ProjectSettings::get_singleton()->set("importer_defaults/" + params->importer->get_importer_name(), d);
+			for (const PropertyInfo &E : params->properties) {
+				import_settings[E.name] = params->values[E.name];
+			}
+
+			ProjectSettings::get_singleton()->set(setting_name, import_settings);
 			ProjectSettings::get_singleton()->save();
 			_update_preset_menu();
 		} break;
 		case ITEM_LOAD_DEFAULT: {
-			ERR_FAIL_COND(!ProjectSettings::get_singleton()->has_setting("importer_defaults/" + params->importer->get_importer_name()));
+			ERR_FAIL_COND(!ProjectSettings::get_singleton()->has_setting(setting_name));
 
-			Dictionary d = ProjectSettings::get_singleton()->get("importer_defaults/" + params->importer->get_importer_name());
-			List<Variant> v;
-			d.get_key_list(&v);
+			Dictionary import_settings = GLOBAL_GET(setting_name);
+			List<Variant> keys;
+			import_settings.get_key_list(&keys);
 
 			if (params->checking) {
 				params->checked.clear();
 			}
-			for (List<Variant>::Element *E = v.front(); E; E = E->next()) {
-				params->values[E->get()] = d[E->get()];
+			for (const Variant &E : keys) {
+				params->values[E] = import_settings[E];
 				if (params->checking) {
-					params->checked.insert(E->get());
+					params->checked.insert(E);
 				}
 			}
 			params->update();
 		} break;
 		case ITEM_CLEAR_DEFAULT: {
-			ProjectSettings::get_singleton()->set("importer_defaults/" + params->importer->get_importer_name(), Variant());
+			ProjectSettings::get_singleton()->set(setting_name, Variant());
 			ProjectSettings::get_singleton()->save();
 			_update_preset_menu();
 		} break;
 		default: {
 			List<ResourceImporter::ImportOption> options;
 
-			params->importer->get_import_options(&options, p_idx);
+			params->importer->get_import_options(params->base_options_path, &options, p_idx);
 
 			if (params->checking) {
 				params->checked.clear();
 			}
-			for (List<ResourceImporter::ImportOption>::Element *E = options.front(); E; E = E->next()) {
-				params->values[E->get().option.name] = E->get().default_value;
+			for (const ResourceImporter::ImportOption &E : options) {
+				params->values[E.option.name] = E.default_value;
 				if (params->checking) {
-					params->checked.insert(E->get().option.name);
+					params->checked.insert(E.option.name);
 				}
 			}
 			params->update();
@@ -419,58 +521,123 @@ static bool _find_owners(EditorFileSystemDirectory *efsd, const String &p_path)
 
 	for (int i = 0; i < efsd->get_file_count(); i++) {
 		Vector<String> deps = efsd->get_file_deps(i);
-		if (deps.find(p_path) != -1) {
+		if (deps.has(p_path)) {
 			return true;
 		}
 	}
 
 	return false;
 }
+
+void ImportDock::_reimport_pressed() {
+	_reimport_attempt();
+
+	if (params->importer.is_valid() && params->paths.size() == 1 && params->importer->has_advanced_options()) {
+		advanced->show();
+		advanced_spacer->show();
+	} else {
+		advanced->hide();
+		advanced_spacer->hide();
+	}
+}
+
 void ImportDock::_reimport_attempt() {
-	bool need_restart = false;
 	bool used_in_resources = false;
 
 	String importer_name;
 	if (params->importer.is_valid()) {
 		importer_name = params->importer->get_importer_name();
 	} else {
-		importer_name = "keep";
+		if (params->skip) {
+			importer_name = "skip";
+		} else {
+			importer_name = "keep";
+		}
 	}
 	for (int i = 0; i < params->paths.size(); i++) {
 		Ref<ConfigFile> config;
-		config.instance();
+		config.instantiate();
 		Error err = config->load(params->paths[i] + ".import");
 		ERR_CONTINUE(err != OK);
 
 		String imported_with = config->get_value("remap", "importer");
-		if (imported_with != importer_name) {
-			need_restart = true;
-			if (_find_owners(EditorFileSystem::get_singleton()->get_filesystem(), params->paths[i])) {
-				used_in_resources = true;
+		if (imported_with != importer_name && imported_with != "keep" && imported_with != "skip") {
+			Ref<Resource> resource = ResourceLoader::load(params->paths[i]);
+			if (resource.is_valid()) {
+				need_cleanup.push_back(params->paths[i]);
+				if (_find_owners(EditorFileSystem::get_singleton()->get_filesystem(), params->paths[i])) {
+					used_in_resources = true;
+				}
 			}
 		}
 	}
 
-	if (need_restart) {
+	if (!need_cleanup.is_empty() || used_in_resources) {
+		cleanup_warning->set_visible(!need_cleanup.is_empty());
 		label_warning->set_visible(used_in_resources);
-		reimport_confirm->popup_centered_minsize();
+		reimport_confirm->popup_centered();
 		return;
 	}
 
 	_reimport();
 }
 
-void ImportDock::_reimport_and_restart() {
-	EditorNode::get_singleton()->save_all_scenes();
-	EditorResourcePreview::get_singleton()->stop(); //don't try to re-create previews after import
+void ImportDock::_reimport_and_cleanup() {
+	HashMap<String, Ref<Resource>> old_resources;
+
+	for (const String &path : need_cleanup) {
+		Ref<Resource> res = ResourceLoader::load(path);
+		res->set_path("");
+		res->set_meta(SNAME("_skip_save_"), true);
+		old_resources[path] = res;
+	}
+
+	EditorResourcePreview::get_singleton()->stop(); // Don't try to re-create previews after import.
 	_reimport();
-	EditorNode::get_singleton()->restart_editor();
+
+	if (need_cleanup.is_empty()) {
+		return;
+	}
+
+	// After changing resource type we need to make sure that all old instances are unloaded or replaced.
+	EditorNode::get_singleton()->push_item(nullptr);
+	EditorUndoRedoManager::get_singleton()->clear_history();
+
+	List<Ref<Resource>> external_resources;
+	ResourceCache::get_cached_resources(&external_resources);
+
+	Vector<Ref<Resource>> old_resources_to_replace;
+	Vector<Ref<Resource>> new_resources_to_replace;
+	for (const String &path : need_cleanup) {
+		Ref<Resource> old_res = old_resources[path];
+		if (params->importer.is_valid()) {
+			Ref<Resource> new_res = ResourceLoader::load(path);
+			if (new_res.is_valid()) {
+				old_resources_to_replace.append(old_res);
+				new_resources_to_replace.append(new_res);
+			}
+		}
+	}
+
+	EditorNode::get_singleton()->replace_resources_in_scenes(old_resources_to_replace, new_resources_to_replace);
+
+	for (Ref<Resource> res : external_resources) {
+		EditorNode::get_singleton()->replace_resources_in_object(res.ptr(), old_resources_to_replace, new_resources_to_replace);
+	}
+
+	need_cleanup.clear();
+}
+
+void ImportDock::_advanced_options() {
+	if (params->paths.size() == 1 && params->importer.is_valid()) {
+		params->importer->show_advanced_options(params->paths[0]);
+	}
 }
 
 void ImportDock::_reimport() {
 	for (int i = 0; i < params->paths.size(); i++) {
 		Ref<ConfigFile> config;
-		config.instance();
+		config.instantiate();
 		Error err = config->load(params->paths[i] + ".import");
 		ERR_CONTINUE(err != OK);
 
@@ -479,9 +646,9 @@ void ImportDock::_reimport() {
 
 			if (params->checking && config->get_value("remap", "importer") == params->importer->get_importer_name()) {
 				//update only what is edited (checkboxes) if the importer is the same
-				for (List<PropertyInfo>::Element *E = params->properties.front(); E; E = E->next()) {
-					if (params->checked.has(E->get().name)) {
-						config->set_value("params", E->get().name, params->values[E->get().name]);
+				for (const PropertyInfo &E : params->properties) {
+					if (params->checked.has(E.name)) {
+						config->set_value("params", E.name, params->values[E.name]);
 					}
 				}
 			} else {
@@ -491,16 +658,16 @@ void ImportDock::_reimport() {
 					config->erase_section("params");
 				}
 
-				for (List<PropertyInfo>::Element *E = params->properties.front(); E; E = E->next()) {
-					config->set_value("params", E->get().name, params->values[E->get().name]);
+				for (const PropertyInfo &E : params->properties) {
+					config->set_value("params", E.name, params->values[E.name]);
 				}
 			}
 
 			//handle group file
 			Ref<ResourceImporter> importer = ResourceFormatImporter::get_singleton()->get_importer_by_name(importer_name);
-			ERR_CONTINUE(!importer.is_valid());
+			ERR_CONTINUE(importer.is_null());
 			String group_file_property = importer->get_option_group_file();
-			if (group_file_property != String()) {
+			if (!group_file_property.is_empty()) {
 				//can import from a group (as in, atlas)
 				ERR_CONTINUE(!params->values.has(group_file_property));
 				String group_file = params->values[group_file_property];
@@ -512,14 +679,18 @@ void ImportDock::_reimport() {
 		} else {
 			//set to no import
 			config->clear();
-			config->set_value("remap", "importer", "keep");
+			if (params->skip) {
+				config->set_value("remap", "importer", "skip");
+			} else {
+				config->set_value("remap", "importer", "keep");
+			}
 		}
 
 		config->save(params->paths[i] + ".import");
 	}
 
 	EditorFileSystem::get_singleton()->reimport_files(params->paths);
-	EditorFileSystem::get_singleton()->emit_signal("filesystem_changed"); //it changed, so force emitting the signal
+	EditorFileSystem::get_singleton()->emit_signal(SNAME("filesystem_changed")); //it changed, so force emitting the signal
 
 	_set_dirty(false);
 }
@@ -527,13 +698,14 @@ void ImportDock::_reimport() {
 void ImportDock::_notification(int p_what) {
 	switch (p_what) {
 		case EditorSettings::NOTIFICATION_EDITOR_SETTINGS_CHANGED: {
-			imported->add_style_override("normal", get_stylebox("normal", "LineEdit"));
-			import_opts->set_property_name_style(EditorPropertyNameProcessor::get_settings_style());
+			if (EditorThemeManager::is_generated_theme_outdated()) {
+				imported->add_theme_style_override(CoreStringName(normal), get_theme_stylebox(CoreStringName(normal), SNAME("LineEdit")));
+			}
 		} break;
 
 		case NOTIFICATION_ENTER_TREE: {
 			import_opts->edit(params);
-			label_warning->add_color_override("font_color", get_color("warning_color", "Editor"));
+			label_warning->add_theme_color_override(SceneStringName(font_color), get_theme_color(SNAME("warning_color"), EditorStringName(Editor)));
 		} break;
 	}
 }
@@ -546,13 +718,13 @@ void ImportDock::_set_dirty(bool p_dirty) {
 	if (p_dirty) {
 		// Add a dirty marker to notify the user that they should reimport the selected resource to see changes.
 		import->set_text(TTR("Reimport") + " (*)");
-		import->add_color_override("font_color", get_color("warning_color", "Editor"));
-		import->set_tooltip(TTR("You have pending changes that haven't been applied yet. Click Reimport to apply changes made to the import options.\nSelecting another resource in the FileSystem dock without clicking Reimport first will discard changes made in the Import dock."));
+		import->add_theme_color_override(SceneStringName(font_color), get_theme_color(SNAME("warning_color"), EditorStringName(Editor)));
+		import->set_tooltip_text(TTR("You have pending changes that haven't been applied yet. Click Reimport to apply changes made to the import options.\nSelecting another resource in the FileSystem dock without clicking Reimport first will discard changes made in the Import dock."));
 	} else {
 		// Remove the dirty marker on the Reimport button.
 		import->set_text(TTR("Reimport"));
-		import->add_color_override("font_color", get_color("font_color", "Editor"));
-		import->set_tooltip("");
+		import->remove_theme_color_override(SceneStringName(font_color));
+		import->set_tooltip_text("");
 	}
 }
 
@@ -563,14 +735,9 @@ void ImportDock::_property_toggled(const StringName &p_prop, bool p_checked) {
 		params->checked.erase(p_prop);
 	}
 }
+
 void ImportDock::_bind_methods() {
 	ClassDB::bind_method(D_METHOD("_reimport"), &ImportDock::_reimport);
-	ClassDB::bind_method(D_METHOD("_preset_selected"), &ImportDock::_preset_selected);
-	ClassDB::bind_method(D_METHOD("_importer_selected"), &ImportDock::_importer_selected);
-	ClassDB::bind_method(D_METHOD("_property_edited"), &ImportDock::_property_edited);
-	ClassDB::bind_method(D_METHOD("_property_toggled"), &ImportDock::_property_toggled);
-	ClassDB::bind_method(D_METHOD("_reimport_and_restart"), &ImportDock::_reimport_and_restart);
-	ClassDB::bind_method(D_METHOD("_reimport_attempt"), &ImportDock::_reimport_attempt);
 }
 
 void ImportDock::initialize_import_options() const {
@@ -580,6 +747,7 @@ void ImportDock::initialize_import_options() const {
 }
 
 ImportDock::ImportDock() {
+	singleton = this;
 	set_name("Import");
 
 	content = memnew(VBoxContainer);
@@ -588,47 +756,60 @@ ImportDock::ImportDock() {
 	content->hide();
 
 	imported = memnew(Label);
-	imported->add_style_override("normal", EditorNode::get_singleton()->get_gui_base()->get_stylebox("normal", "LineEdit"));
+	imported->add_theme_style_override(CoreStringName(normal), EditorNode::get_singleton()->get_editor_theme()->get_stylebox(CoreStringName(normal), SNAME("LineEdit")));
 	imported->set_clip_text(true);
 	content->add_child(imported);
 	HBoxContainer *hb = memnew(HBoxContainer);
 	content->add_margin_child(TTR("Import As:"), hb);
 	import_as = memnew(OptionButton);
 	import_as->set_disabled(true);
-	import_as->connect("item_selected", this, "_importer_selected");
+	import_as->set_fit_to_longest_item(false);
+	import_as->set_text_overrun_behavior(TextServer::OVERRUN_TRIM_ELLIPSIS);
+	import_as->set_h_size_flags(SIZE_EXPAND_FILL);
+	import_as->connect(SceneStringName(item_selected), callable_mp(this, &ImportDock::_importer_selected));
 	hb->add_child(import_as);
 	import_as->set_h_size_flags(SIZE_EXPAND_FILL);
 	preset = memnew(MenuButton);
 	preset->set_text(TTR("Preset"));
 	preset->set_disabled(true);
-	preset->get_popup()->connect("index_pressed", this, "_preset_selected");
+	preset->get_popup()->connect("index_pressed", callable_mp(this, &ImportDock::_preset_selected));
 	hb->add_child(preset);
 
 	import_opts = memnew(EditorInspector);
 	content->add_child(import_opts);
 	import_opts->set_v_size_flags(SIZE_EXPAND_FILL);
-	import_opts->set_property_name_style(EditorPropertyNameProcessor::get_settings_style());
-	import_opts->connect("property_edited", this, "_property_edited");
-	import_opts->connect("property_toggled", this, "_property_toggled");
+	import_opts->connect("property_edited", callable_mp(this, &ImportDock::_property_edited));
+	import_opts->connect("property_toggled", callable_mp(this, &ImportDock::_property_toggled));
+	// Make it possible to display tooltips stored in the XML class reference.
+	// The object name is set when the importer changes in `_update_options()`.
+	import_opts->set_use_doc_hints(true);
 
 	hb = memnew(HBoxContainer);
 	content->add_child(hb);
 	import = memnew(Button);
 	import->set_text(TTR("Reimport"));
 	import->set_disabled(true);
-	import->connect("pressed", this, "_reimport_attempt");
+	import->connect(SceneStringName(pressed), callable_mp(this, &ImportDock::_reimport_pressed));
+	advanced_spacer = hb->add_spacer();
+	advanced = memnew(Button);
+	advanced->set_text(TTR("Advanced..."));
+	hb->add_child(advanced);
 	hb->add_spacer();
 	hb->add_child(import);
 	hb->add_spacer();
 
+	advanced->hide();
+	advanced_spacer->hide();
+	advanced->connect(SceneStringName(pressed), callable_mp(this, &ImportDock::_advanced_options));
+
 	reimport_confirm = memnew(ConfirmationDialog);
-	reimport_confirm->get_ok()->set_text(TTR("Save Scenes, Re-Import, and Restart"));
 	content->add_child(reimport_confirm);
-	reimport_confirm->connect("confirmed", this, "_reimport_and_restart");
+	reimport_confirm->connect(SceneStringName(confirmed), callable_mp(this, &ImportDock::_reimport_and_cleanup));
 
 	VBoxContainer *vbc_confirm = memnew(VBoxContainer());
-	vbc_confirm->add_child(memnew(Label(TTR("Changing the type of an imported file requires editor restart."))));
-	label_warning = memnew(Label(TTR("WARNING: Assets exist that use this resource, they may stop loading properly.")));
+	cleanup_warning = memnew(Label(TTR("The imported resource is currently loaded. All instances will be replaced and undo history will be cleared.")));
+	vbc_confirm->add_child(cleanup_warning);
+	label_warning = memnew(Label(TTR("WARNING: Assets exist that use this resource. They may stop loading properly after changing type.")));
 	vbc_confirm->add_child(label_warning);
 	reimport_confirm->add_child(vbc_confirm);
 
@@ -636,14 +817,15 @@ ImportDock::ImportDock() {
 
 	select_a_resource = memnew(Label);
 	select_a_resource->set_text(TTR("Select a resource file in the filesystem or in the inspector to adjust import settings."));
-	select_a_resource->set_autowrap(true);
+	select_a_resource->set_autowrap_mode(TextServer::AUTOWRAP_WORD);
 	select_a_resource->set_custom_minimum_size(Size2(100 * EDSCALE, 0));
 	select_a_resource->set_v_size_flags(SIZE_EXPAND_FILL);
-	select_a_resource->set_align(Label::ALIGN_CENTER);
-	select_a_resource->set_valign(Label::VALIGN_CENTER);
+	select_a_resource->set_horizontal_alignment(HORIZONTAL_ALIGNMENT_CENTER);
+	select_a_resource->set_vertical_alignment(VERTICAL_ALIGNMENT_CENTER);
 	add_child(select_a_resource);
 }
 
 ImportDock::~ImportDock() {
+	singleton = nullptr;
 	memdelete(params);
 }
