@@ -3,5 +3,5 @@ using System;
 namespace Godot
 {
     [AttributeUsage(AttributeTargets.Delegate)]
-    public class SignalAttribute : Attribute { }
+    public sealed class SignalAttribute : Attribute { }
 }
